@@ -1,23 +1,28 @@
 /*
     FUSE: Filesystem in Userspace
-    Copyright (C) 2001-2005  Miklos Szeredi <miklos@szeredi.hu>
+    Copyright (C) 2001-2006  Miklos Szeredi <miklos@szeredi.hu>
 
     This program can be distributed under the terms of the GNU GPL.
     See the file COPYING.
+
+    gcc -Wall `pkg-config fuse --cflags --libs` fusexmp_fh.c -o fusexmp_fh
 */
 
+#define FUSE_USE_VERSION 26
+
 #include <config.h>
 
 #define _GNU_SOURCE
 
 #include <fuse.h>
+#include <ulockmgr.h>
 #include <stdio.h>
 #include <string.h>
 #include <unistd.h>
 #include <fcntl.h>
 #include <dirent.h>
 #include <errno.h>
-#include <sys/statfs.h>
+#include <sys/time.h>
 #ifdef HAVE_SETXATTR
 #include <sys/xattr.h>
 #endif
@@ -27,7 +32,32 @@ static int xmp_getattr(const char *path, struct stat *stbuf)
     int res;
 
     res = lstat(path, stbuf);
-    if(res == -1)
+    if (res == -1)
+        return -errno;
+
+    return 0;
+}
+
+static int xmp_fgetattr(const char *path, struct stat *stbuf,
+                        struct fuse_file_info *fi)
+{
+    int res;
+
+    (void) path;
+
+    res = fstat(fi->fh, stbuf);
+    if (res == -1)
+        return -errno;
+
+    return 0;
+}
+
+static int xmp_access(const char *path, int mask)
+{
+    int res;
+
+    res = access(path, mask);
+    if (res == -1)
         return -errno;
 
     return 0;
@@ -38,7 +68,7 @@ static int xmp_readlink(const char *path, char *buf, size_t size)
     int res;
 
     res = readlink(path, buf, size - 1);
-    if(res == -1)
+    if (res == -1)
         return -errno;
 
     buf[res] = '\0';
@@ -55,10 +85,15 @@ static int xmp_opendir(const char *path, struct fuse_file_info *fi)
     return 0;
 }
 
+static inline DIR *get_dirp(struct fuse_file_info *fi)
+{
+    return (DIR *) (uintptr_t) fi->fh;
+}
+
 static int xmp_readdir(const char *path, void *buf, fuse_fill_dir_t filler,
                        off_t offset, struct fuse_file_info *fi)
 {
-    DIR *dp = (DIR *) fi->fh;
+    DIR *dp = get_dirp(fi);
     struct dirent *de;
 
     (void) path;
@@ -68,7 +103,7 @@ static int xmp_readdir(const char *path, void *buf, fuse_fill_dir_t filler,
         memset(&st, 0, sizeof(st));
         st.st_ino = de->d_ino;
         st.st_mode = de->d_type << 12;
-        if (filler(buf, de->d_name, &st, de->d_off))
+        if (filler(buf, de->d_name, &st, telldir(dp)))
             break;
     }
 
@@ -77,7 +112,7 @@ static int xmp_readdir(const char *path, void *buf, fuse_fill_dir_t filler,
 
 static int xmp_releasedir(const char *path, struct fuse_file_info *fi)
 {
-    DIR *dp = (DIR *) fi->fh;
+    DIR *dp = get_dirp(fi);
     (void) path;
     closedir(dp);
     return 0;
@@ -87,8 +122,11 @@ static int xmp_mknod(const char *path, mode_t mode, dev_t rdev)
 {
     int res;
 
-    res = mknod(path, mode, rdev);
-    if(res == -1)
+    if (S_ISFIFO(mode))
+        res = mkfifo(path, mode);
+    else
+        res = mknod(path, mode, rdev);
+    if (res == -1)
         return -errno;
 
     return 0;
@@ -99,7 +137,7 @@ static int xmp_mkdir(const char *path, mode_t mode)
     int res;
 
     res = mkdir(path, mode);
-    if(res == -1)
+    if (res == -1)
         return -errno;
 
     return 0;
@@ -110,7 +148,7 @@ static int xmp_unlink(const char *path)
     int res;
 
     res = unlink(path);
-    if(res == -1)
+    if (res == -1)
         return -errno;
 
     return 0;
@@ -121,7 +159,7 @@ static int xmp_rmdir(const char *path)
     int res;
 
     res = rmdir(path);
-    if(res == -1)
+    if (res == -1)
         return -errno;
 
     return 0;
@@ -132,7 +170,7 @@ static int xmp_symlink(const char *from, const char *to)
     int res;
 
     res = symlink(from, to);
-    if(res == -1)
+    if (res == -1)
         return -errno;
 
     return 0;
@@ -143,7 +181,7 @@ static int xmp_rename(const char *from, const char *to)
     int res;
 
     res = rename(from, to);
-    if(res == -1)
+    if (res == -1)
         return -errno;
 
     return 0;
@@ -154,7 +192,7 @@ static int xmp_link(const char *from, const char *to)
     int res;
 
     res = link(from, to);
-    if(res == -1)
+    if (res == -1)
         return -errno;
 
     return 0;
@@ -165,7 +203,7 @@ static int xmp_chmod(const char *path, mode_t mode)
     int res;
 
     res = chmod(path, mode);
-    if(res == -1)
+    if (res == -1)
         return -errno;
 
     return 0;
@@ -176,7 +214,7 @@ static int xmp_chown(const char *path, uid_t uid, gid_t gid)
     int res;
 
     res = lchown(path, uid, gid);
-    if(res == -1)
+    if (res == -1)
         return -errno;
 
     return 0;
@@ -187,30 +225,61 @@ static int xmp_truncate(const char *path, off_t size)
     int res;
 
     res = truncate(path, size);
-    if(res == -1)
+    if (res == -1)
         return -errno;
 
     return 0;
 }
 
-static int xmp_utime(const char *path, struct utimbuf *buf)
+static int xmp_ftruncate(const char *path, off_t size,
+                         struct fuse_file_info *fi)
 {
     int res;
 
-    res = utime(path, buf);
-    if(res == -1)
+    (void) path;
+
+    res = ftruncate(fi->fh, size);
+    if (res == -1)
         return -errno;
 
     return 0;
 }
 
+static int xmp_utimens(const char *path, const struct timespec ts[2])
+{
+    int res;
+    struct timeval tv[2];
+
+    tv[0].tv_sec = ts[0].tv_sec;
+    tv[0].tv_usec = ts[0].tv_nsec / 1000;
+    tv[1].tv_sec = ts[1].tv_sec;
+    tv[1].tv_usec = ts[1].tv_nsec / 1000;
+
+    res = utimes(path, tv);
+    if (res == -1)
+        return -errno;
+
+    return 0;
+}
+
+static int xmp_create(const char *path, mode_t mode, struct fuse_file_info *fi)
+{
+    int fd;
+
+    fd = open(path, fi->flags, mode);
+    if (fd == -1)
+        return -errno;
+
+    fi->fh = fd;
+    return 0;
+}
 
 static int xmp_open(const char *path, struct fuse_file_info *fi)
 {
     int fd;
 
     fd = open(path, fi->flags);
-    if(fd == -1)
+    if (fd == -1)
         return -errno;
 
     fi->fh = fd;
@@ -224,7 +293,7 @@ static int xmp_read(const char *path, char *buf, size_t size, off_t offset,
 
     (void) path;
     res = pread(fi->fh, buf, size, offset);
-    if(res == -1)
+    if (res == -1)
         res = -errno;
 
     return res;
@@ -237,18 +306,35 @@ static int xmp_write(const char *path, const char *buf, size_t size,
 
     (void) path;
     res = pwrite(fi->fh, buf, size, offset);
-    if(res == -1)
+    if (res == -1)
         res = -errno;
 
     return res;
 }
 
-static int xmp_statfs(const char *path, struct statfs *stbuf)
+static int xmp_statfs(const char *path, struct statvfs *stbuf)
 {
     int res;
 
-    res = statfs(path, stbuf);
-    if(res == -1)
+    res = statvfs(path, stbuf);
+    if (res == -1)
+        return -errno;
+
+    return 0;
+}
+
+static int xmp_flush(const char *path, struct fuse_file_info *fi)
+{
+    int res;
+
+    (void) path;
+    /* This is called from every close on an open file, so call the
+       close on the underlying filesystem.  But since flush may be
+       called multiple times for an open file, this must not really
+       close the file.  This is important if used on a network
+       filesystem like NFS which flush the data/metadata on close() */
+    res = close(dup(fi->fh));
+    if (res == -1)
         return -errno;
 
     return 0;
@@ -268,11 +354,15 @@ static int xmp_fsync(const char *path, int isdatasync,
     int res;
     (void) path;
 
+#ifndef HAVE_FDATASYNC
+    (void) isdatasync;
+#else
     if (isdatasync)
         res = fdatasync(fi->fh);
     else
+#endif
         res = fsync(fi->fh);
-    if(res == -1)
+    if (res == -1)
         return -errno;
 
     return 0;
@@ -284,7 +374,7 @@ static int xmp_setxattr(const char *path, const char *name, const char *value,
                         size_t size, int flags)
 {
     int res = lsetxattr(path, name, value, size, flags);
-    if(res == -1)
+    if (res == -1)
         return -errno;
     return 0;
 }
@@ -293,7 +383,7 @@ static int xmp_getxattr(const char *path, const char *name, char *value,
                     size_t size)
 {
     int res = lgetxattr(path, name, value, size);
-    if(res == -1)
+    if (res == -1)
         return -errno;
     return res;
 }
@@ -301,7 +391,7 @@ static int xmp_getxattr(const char *path, const char *name, char *value,
 static int xmp_listxattr(const char *path, char *list, size_t size)
 {
     int res = llistxattr(path, list, size);
-    if(res == -1)
+    if (res == -1)
         return -errno;
     return res;
 }
@@ -309,14 +399,25 @@ static int xmp_listxattr(const char *path, char *list, size_t size)
 static int xmp_removexattr(const char *path, const char *name)
 {
     int res = lremovexattr(path, name);
-    if(res == -1)
+    if (res == -1)
         return -errno;
     return 0;
 }
 #endif /* HAVE_SETXATTR */
 
+static int xmp_lock(const char *path, struct fuse_file_info *fi, int cmd,
+                    struct flock *lock)
+{
+    (void) path;
+
+    return ulockmgr_op(fi->fh, cmd, lock, &fi->lock_owner,
+                       sizeof(fi->lock_owner));
+}
+
 static struct fuse_operations xmp_oper = {
     .getattr	= xmp_getattr,
+    .fgetattr	= xmp_fgetattr,
+    .access	= xmp_access,
     .readlink	= xmp_readlink,
     .opendir	= xmp_opendir,
     .readdir	= xmp_readdir,
@@ -331,11 +432,14 @@ static struct fuse_operations xmp_oper = {
     .chmod	= xmp_chmod,
     .chown	= xmp_chown,
     .truncate	= xmp_truncate,
-    .utime	= xmp_utime,
+    .ftruncate	= xmp_ftruncate,
+    .utimens	= xmp_utimens,
+    .create	= xmp_create,
     .open	= xmp_open,
     .read	= xmp_read,
     .write	= xmp_write,
     .statfs	= xmp_statfs,
+    .flush	= xmp_flush,
     .release	= xmp_release,
     .fsync	= xmp_fsync,
 #ifdef HAVE_SETXATTR
@@ -344,10 +448,11 @@ static struct fuse_operations xmp_oper = {
     .listxattr	= xmp_listxattr,
     .removexattr= xmp_removexattr,
 #endif
+    .lock	= xmp_lock,
 };
 
 int main(int argc, char *argv[])
 {
     umask(0);
-    return fuse_main(argc, argv, &xmp_oper);
+    return fuse_main(argc, argv, &xmp_oper, NULL);
 }
