@@ -1,6 +1,6 @@
 /*
     FUSE: Filesystem in Userspace
-    Copyright (C) 2001-2005  Miklos Szeredi <miklos@szeredi.hu>
+    Copyright (C) 2001-2006  Miklos Szeredi <miklos@szeredi.hu>
 
     This program can be distributed under the terms of the GNU LGPL.
     See the file COPYING.LIB.
@@ -13,12 +13,22 @@
  * Low level API                                               *
  * =========================================================== */
 
+/* IMPORTANT: you should define FUSE_USE_VERSION before including this
+   header.  To use the newest API define it to 26 (recommended for any
+   new application), to use the old API define it to 24 (default) or
+   25 */
+
+#ifndef FUSE_USE_VERSION
+#define FUSE_USE_VERSION 24
+#endif
+
 #include "fuse_common.h"
 
 #include <utime.h>
+#include <fcntl.h>
 #include <sys/types.h>
 #include <sys/stat.h>
-#include <sys/statfs.h>
+#include <sys/statvfs.h>
 #include <sys/uio.h>
 
 #ifdef __cplusplus
@@ -55,7 +65,13 @@ struct fuse_chan;
 
 /** Directory entry parameters supplied to fuse_reply_entry() */
 struct fuse_entry_param {
-    /** Unique inode number */
+    /** Unique inode number
+     *
+     * In lookup, zero means negative entry (from version 2.5)
+     * Returning ENOENT also means negative entry, but by setting zero
+     * ino the kernel may cache negative entries for entry_timeout
+     * seconds.
+     */
     fuse_ino_t ino;
 
     /** The ino/generation pair should be unique for the filesystem's
@@ -127,7 +143,7 @@ struct fuse_lowlevel_ops {
      *
      * @param userdata the user data passed to fuse_lowlevel_new()
      */
-    void (*init) (void *userdata);
+    void (*init) (void *userdata, struct fuse_conn_info *conn);
 
     /**
      * Clean up filesystem
@@ -144,8 +160,8 @@ struct fuse_lowlevel_ops {
      * Look up a directory entry by name
      *
      * Valid replies:
-     *   fuse_reply_entry()
-     *   fuse_reply_err()
+     *   fuse_reply_entry
+     *   fuse_reply_err
      *
      * @param req request handle
      * @param parent inode number of the parent directory
@@ -170,7 +186,7 @@ struct fuse_lowlevel_ops {
      * will receive a forget message.
      *
      * Valid replies:
-     *   fuse_reply_none()
+     *   fuse_reply_none
      *
      * @param req request handle
      * @param ino the inode number
@@ -182,8 +198,8 @@ struct fuse_lowlevel_ops {
      * Get file attributes
      *
      * Valid replies:
-     *   fuse_reply_attr()
-     *   fuse_reply_err()
+     *   fuse_reply_attr
+     *   fuse_reply_err
      *
      * @param req request handle
      * @param ino the inode number
@@ -198,15 +214,25 @@ struct fuse_lowlevel_ops {
      * bitmask contain valid values.  Other members contain undefined
      * values.
      *
+     * If the setattr was invoked from the ftruncate() system call
+     * under Linux kernel versions 2.6.15 or later, the fi->fh will
+     * contain the value set by the open method or will be undefined
+     * if the open method didn't set any value.  Otherwise (not
+     * ftruncate call, or kernel version earlier than 2.6.15) the fi
+     * parameter will be NULL.
+     *
      * Valid replies:
-     *   fuse_reply_attr()
-     *   fuse_reply_err()
+     *   fuse_reply_attr
+     *   fuse_reply_err
      *
      * @param req request handle
      * @param ino the inode number
      * @param attr the attributes
      * @param to_set bit mask of attributes which should be set
-     * @param fi for future use, currently always NULL
+     * @param fi file information, or NULL
+     *
+     * Changed in version 2.5:
+     *     file information filled in for ftruncate
      */
     void (*setattr) (fuse_req_t req, fuse_ino_t ino, struct stat *attr,
                     int to_set, struct fuse_file_info *fi);
@@ -423,6 +449,9 @@ struct fuse_lowlevel_ops {
      * One reason to flush data, is if the filesystem wants to return
      * write errors.
      *
+     * If the filesystem supports file locking operations (setlk,
+     * getlk) it should remove all locks belonging to 'fi->owner'.
+     *
      * Valid replies:
      *   fuse_reply_err
      *
@@ -501,7 +530,7 @@ struct fuse_lowlevel_ops {
     /**
      * Read directory
      *
-     * Send a buffer filled using fuse_add_dirent(), with size not
+     * Send a buffer filled using fuse_add_direntry(), with size not
      * exceeding the requested size.  Send an empty buffer on end of
      * stream.
      *
@@ -570,8 +599,9 @@ struct fuse_lowlevel_ops {
      *   fuse_reply_err
      *
      * @param req request handle
+     * @param ino the inode number, zero means "undefined"
      */
-    void (*statfs) (fuse_req_t req);
+    void (*statfs) (fuse_req_t req, fuse_ino_t ino);
 
     /**
      * Set an extended attribute
@@ -642,6 +672,126 @@ struct fuse_lowlevel_ops {
      * @param name of the extended attribute
      */
     void (*removexattr) (fuse_req_t req, fuse_ino_t ino, const char *name);
+
+    /**
+     * Check file access permissions
+     *
+     * This will be called for the access() system call.  If the
+     * 'default_permissions' mount option is given, this method is not
+     * called.
+     *
+     * This method is not called under Linux kernel versions 2.4.x
+     *
+     * Introduced in version 2.5
+     *
+     * Valid replies:
+     *   fuse_reply_err
+     *
+     * @param req request handle
+     * @param ino the inode number
+     * @param mask requested access mode
+     */
+    void (*access) (fuse_req_t req, fuse_ino_t ino, int mask);
+
+    /**
+     * Create and open a file
+     *
+     * If the file does not exist, first create it with the specified
+     * mode, and then open it.
+     *
+     * Open flags (with the exception of O_NOCTTY) are available in
+     * fi->flags.
+     *
+     * Filesystem may store an arbitrary file handle (pointer, index,
+     * etc) in fi->fh, and use this in other all other file operations
+     * (read, write, flush, release, fsync).
+     *
+     * There are also some flags (direct_io, keep_cache) which the
+     * filesystem may set in fi, to change the way the file is opened.
+     * See fuse_file_info structure in <fuse_common.h> for more details.
+     *
+     * If this method is not implemented or under Linux kernel
+     * versions earlier than 2.6.15, the mknod() and open() methods
+     * will be called instead.
+     *
+     * Introduced in version 2.5
+     *
+     * Valid replies:
+     *   fuse_reply_create
+     *   fuse_reply_err
+     *
+     * @param req request handle
+     * @param parent inode number of the parent directory
+     * @param name to create
+     * @param mode file type and mode with which to create the new file
+     * @param fi file information
+     */
+    void (*create) (fuse_req_t req, fuse_ino_t parent, const char *name,
+                    mode_t mode, struct fuse_file_info *fi);
+
+    /**
+     * Test for a POSIX file lock
+     *
+     * Introduced in version 2.6
+     *
+     * Valid replies:
+     *   fuse_reply_lock
+     *   fuse_reply_err
+     *
+     * @param req request handle
+     * @param ino the inode number
+     * @param fi file information
+     * @param lock the region/type to test
+     */
+    void (*getlk) (fuse_req_t req, fuse_ino_t ino, struct fuse_file_info *fi,
+                   struct flock *lock);
+
+    /**
+     * Acquire, modify or release a POSIX file lock
+     *
+     * For POSIX threads (NPTL) there's a 1-1 relation between pid and
+     * owner, but otherwise this is not always the case.  For checking
+     * lock ownership, 'fi->owner' must be used.  The l_pid field in
+     * 'struct flock' should only be used to fill in this field in
+     * getlk().
+     *
+     * Note: if the locking methods are not implemented, the kernel
+     * will still allow file locking to work locally.  Hence these are
+     * only interesting for network filesystems and similar.
+     *
+     * Introduced in version 2.6
+     *
+     * Valid replies:
+     *   fuse_reply_err
+     *
+     * @param req request handle
+     * @param ino the inode number
+     * @param fi file information
+     * @param lock the region/type to test
+     * @param sleep locking operation may sleep
+     */
+    void (*setlk) (fuse_req_t req, fuse_ino_t ino, struct fuse_file_info *fi,
+                   struct flock *lock, int sleep);
+
+    /**
+     * Map block index within file to block index within device
+     *
+     * Note: This makes sense only for block device backed filesystems
+     * mounted with the 'blkdev' option
+     *
+     * Introduced in version 2.6
+     *
+     * Valid replies:
+     *   fuse_reply_bmap
+     *   fuse_reply_err
+     *
+     * @param req request handle
+     * @param ino the inode number
+     * @param blocksize unit of block index
+     * @param idx block index within file
+     */
+    void (*bmap) (fuse_req_t req, fuse_ino_t ino, size_t blocksize,
+                  uint64_t idx);
 };
 
 /**
@@ -650,8 +800,8 @@ struct fuse_lowlevel_ops {
  * Possible requests:
  *   all except forget
  *
- * unlink, rmdir, rename, flush, release, fsync, fsyncdir, setxattr
- * and removexattr may send a zero code
+ * unlink, rmdir, rename, flush, release, fsync, fsyncdir, setxattr,
+ * removexattr and setlk may send a zero code
  *
  * @param req request handle
  * @param err the positive error value, or zero for success
@@ -681,6 +831,23 @@ void fuse_reply_none(fuse_req_t req);
  */
 int fuse_reply_entry(fuse_req_t req, const struct fuse_entry_param *e);
 
+/**
+ * Reply with a directory entry and open parameters
+ *
+ * currently the following members of 'fi' are used:
+ *   fh, direct_io, keep_cache
+ *
+ * Possible requests:
+ *   create
+ *
+ * @param req request handle
+ * @param e the entry parameters
+ * @param fi file information
+ * @return zero for success, -errno for failure to send reply
+ */
+int fuse_reply_create(fuse_req_t req, const struct fuse_entry_param *e,
+                      const struct fuse_file_info *fi);
+
 /**
  * Reply with attributes
  *
@@ -757,7 +924,7 @@ int fuse_reply_buf(fuse_req_t req, const char *buf, size_t size);
  * @param stbuf filesystem statistics
  * @return zero for success, -errno for failure to send reply
  */
-int fuse_reply_statfs(fuse_req_t req, const struct statfs *stbuf);
+int fuse_reply_statfs(fuse_req_t req, const struct statvfs *stbuf);
 
 /**
  * Reply with needed buffer size
@@ -771,22 +938,42 @@ int fuse_reply_statfs(fuse_req_t req, const struct statfs *stbuf);
  */
 int fuse_reply_xattr(fuse_req_t req, size_t count);
 
+/**
+ * Reply with file lock information
+ *
+ * Possible requests:
+ *   getlk
+ *
+ * @param req request handle
+ * @param lock the lock information
+ * @return zero for success, -errno for failure to send reply
+ */
+int fuse_reply_lock(fuse_req_t req, struct flock *lock);
+
+/**
+ * Reply with block index
+ *
+ * Possible requests:
+ *   bmap
+ *
+ * @param req request handle
+ * @param idx block index within device 
+ * @return zero for success, -errno for failure to send reply
+ */
+int fuse_reply_bmap(fuse_req_t req, uint64_t idx);
+
 /* ----------------------------------------------------------- *
  * Filling a buffer in readdir                                 *
  * ----------------------------------------------------------- */
 
-/**
- * Calculate the number of bytes a directory entry takes up
- *
- * @param namelen the length of the entry name
- * @return the number of bytes needed
- */
-size_t fuse_dirent_size(size_t namelen);
-
 /**
  * Add a directory entry to the buffer
  *
- * Buffer needs to be large enough to hold the entry
+ * Buffer needs to be large enough to hold the entry.  Of it's not,
+ * then the entry is not filled in but the size of the entry is still
+ * returned.  The caller can check this by comparing the bufsize
+ * parameter with the returned entry size.  If the entry size is
+ * larger than the buffer size, the operation failed.
  *
  * From the 'stbuf' argument the st_ino field and bits 12-15 of the
  * st_mode field are used.  The other fields are ignored.
@@ -795,14 +982,17 @@ size_t fuse_dirent_size(size_t namelen);
  * could be any marker, that enables the implementation to find a
  * specific point in the directory stream.
  *
+ * @param req request handle
  * @param buf the point where the new entry will be added to the buffer
+ * @param bufsize remaining size of the buffer
  * @param the name of the entry
  * @param stbuf the file attributes
  * @param off the offset of the next entry
- * @return a pointer to the start of the next entry in the buffer
+ * @return the space needed for the entry
  */
-char *fuse_add_dirent(char *buf, const char *name, const struct stat *stbuf,
-                      off_t off);
+size_t fuse_add_direntry(fuse_req_t req, char *buf, size_t bufsize,
+                         const char *name, const struct stat *stbuf,
+                         off_t off);
 
 /* ----------------------------------------------------------- *
  * Utility functions                                           *
@@ -827,40 +1017,56 @@ void *fuse_req_userdata(fuse_req_t req);
  */
 const struct fuse_ctx *fuse_req_ctx(fuse_req_t req);
 
+/**
+ * Callback function for an interrupt
+ *
+ * @param req interrupted request
+ * @param data user data
+ */
+typedef void (*fuse_interrupt_func_t)(fuse_req_t req, void *data);
+
+/**
+ * Register/unregister callback for an interrupt
+ *
+ * If an interrupt has already happened, then the callback function is
+ * called from within this function, hence it's not possible for
+ * interrupts to be lost.
+ *
+ * @param req request handle
+ * @param func the callback function or NULL for unregister
+ * @parm data user data passed to the callback function
+ */
+void fuse_req_interrupt_func(fuse_req_t req, fuse_interrupt_func_t func,
+                             void *data);
+
+/**
+ * Check if a request has already been interrupted
+ *
+ * @param req request handle
+ * @return 1 if the request has been interrupted, 0 otherwise
+ */
+int fuse_req_interrupted(fuse_req_t req);
+
 /* ----------------------------------------------------------- *
  * Filesystem setup                                            *
  * ----------------------------------------------------------- */
 
-/**
- * Check whether a mount option is claimed by the low level library or
- * not
- *
- * @param opt the option to check
- * @return 1 if it is a low level library option, 0 otherwise
- */
+/* Deprecated, don't use */
 int fuse_lowlevel_is_lib_option(const char *opt);
 
 /**
  * Create a low level session
  *
- * @param opts the comma separated list of options
+ * @param args argument vector
  * @param op the low level filesystem operations
  * @param op_size sizeof(struct fuse_lowlevel_ops)
  * @param userdata user data
  * @return the created session object, or NULL on failure
  */
-struct fuse_session *fuse_lowlevel_new(const char *opts,
+struct fuse_session *fuse_lowlevel_new(struct fuse_args *args,
                                        const struct fuse_lowlevel_ops *op,
                                        size_t op_size, void *userdata);
 
-/**
- * Create a kernel channel
- *
- * @param fd the file descriptor obtained from fuse_mount()
- * @return the created channel object, or NULL on failure
- */
-struct fuse_chan *fuse_kern_chan_new(int fd);
-
 /* ----------------------------------------------------------- *
  * Session interface                                           *
  * ----------------------------------------------------------- */
@@ -928,6 +1134,15 @@ struct fuse_session *fuse_session_new(struct fuse_session_ops *op, void *data);
  */
 void fuse_session_add_chan(struct fuse_session *se, struct fuse_chan *ch);
 
+/**
+ * Remove a channel from a session
+ *
+ * If the channel is not assigned to a session, then this is a no-op
+ *
+ * @param ch the channel to remove
+ */
+void fuse_session_remove_chan(struct fuse_chan *ch);
+
 /**
  * Iterate over the channels assigned to a session
  *
@@ -1011,12 +1226,12 @@ struct fuse_chan_ops {
     /**
      * Hook for receiving a raw request
      *
-     * @param ch the channel
+     * @param ch pointer to the channel
      * @param buf the buffer to store the request in
      * @param size the size of the buffer
      * @return the actual size of the raw request, or -1 on error
      */
-    int (*receive)(struct fuse_chan *ch, char *buf, size_t size);
+    int (*receive)(struct fuse_chan **chp, char *buf, size_t size);
 
     /**
      * Hook for sending a raw reply
@@ -1029,8 +1244,7 @@ struct fuse_chan_ops {
      * @param count the number of blocks in vector
      * @return zero on success, -errno on failure
      */
-    int (*send)(struct fuse_chan *ch, const struct iovec iov[],
-                size_t count);
+    int (*send)(struct fuse_chan *ch, const struct iovec iov[], size_t count);
 
     /**
      * Destroy the channel
@@ -1087,12 +1301,14 @@ struct fuse_session *fuse_chan_session(struct fuse_chan *ch);
 /**
  * Receive a raw request
  *
- * @param ch the channel
+ * A return value of -ENODEV means, that the filesystem was unmounted
+ *
+ * @param ch pointer to the channel
  * @param buf the buffer to store the request in
  * @param size the size of the buffer
- * @return the actual size of the raw request, or -1 on error
+ * @return the actual size of the raw request, or -errno on error
  */
-int fuse_chan_receive(struct fuse_chan *ch, char *buf, size_t size);
+int fuse_chan_recv(struct fuse_chan **ch, char *buf, size_t size);
 
 /**
  * Send a raw reply
@@ -1115,6 +1331,28 @@ int fuse_chan_send(struct fuse_chan *ch, const struct iovec iov[],
  */
 void fuse_chan_destroy(struct fuse_chan *ch);
 
+/* ----------------------------------------------------------- *
+ * Compatibility stuff                                         *
+ * ----------------------------------------------------------- */
+
+#if FUSE_USE_VERSION < 26
+#  include "fuse_lowlevel_compat.h"
+#  define fuse_chan_ops fuse_chan_ops_compat24
+#  define fuse_chan_new fuse_chan_new_compat24
+#  if FUSE_USE_VERSION == 25
+#    define fuse_lowlevel_ops fuse_lowlevel_ops_compat25
+#    define fuse_lowlevel_new fuse_lowlevel_new_compat25
+#  elif FUSE_USE_VERSION == 24
+#    define fuse_lowlevel_ops fuse_lowlevel_ops_compat
+#    define fuse_lowlevel_new fuse_lowlevel_new_compat
+#    define fuse_file_info fuse_file_info_compat
+#    define fuse_reply_statfs fuse_reply_statfs_compat
+#    define fuse_reply_open fuse_reply_open_compat
+#  else
+#    error Compatibility with low-level API version < 24 not supported
+#  endif
+#endif
+
 #ifdef __cplusplus
 }
 #endif
