@@ -10,6 +10,7 @@
 #include "fuse_i.h"
 #include "fuse_opt.h"
 #include "fuse_lowlevel.h"
+#include "fuse_common_compat.h"
 
 #include <stdio.h>
 #include <stdlib.h>
@@ -17,12 +18,12 @@
 #include <unistd.h>
 #include <string.h>
 #include <limits.h>
+#include <errno.h>
 
 enum  {
     KEY_HELP,
     KEY_HELP_NOHEADER,
     KEY_VERSION,
-    KEY_KEEP,
 };
 
 struct helper_opts {
@@ -46,9 +47,9 @@ static const struct fuse_opt fuse_helper_opts[] = {
     FUSE_OPT_KEY("-ho",         KEY_HELP_NOHEADER),
     FUSE_OPT_KEY("-V",          KEY_VERSION),
     FUSE_OPT_KEY("--version",   KEY_VERSION),
-    FUSE_OPT_KEY("-d",          KEY_KEEP),
-    FUSE_OPT_KEY("debug",       KEY_KEEP),
-    FUSE_OPT_KEY("fsname=",     KEY_KEEP),
+    FUSE_OPT_KEY("-d",          FUSE_OPT_KEY_KEEP),
+    FUSE_OPT_KEY("debug",       FUSE_OPT_KEY_KEEP),
+    FUSE_OPT_KEY("fsname=",     FUSE_OPT_KEY_KEEP),
     FUSE_OPT_END
 };
 
@@ -99,15 +100,19 @@ static int fuse_helper_opt_proc(void *data, const char *arg, int key,
         return 1;
 
     case FUSE_OPT_KEY_NONOPT:
-        if (!hopts->mountpoint)
-            return fuse_opt_add_opt(&hopts->mountpoint, arg);
-        else {
+        if (!hopts->mountpoint) {
+            char mountpoint[PATH_MAX];
+            if (realpath(arg, mountpoint) == NULL) {
+                fprintf(stderr, "fuse: bad mount point `%s': %s\n", arg, strerror(errno));
+                return -1;
+            }
+            return fuse_opt_add_opt(&hopts->mountpoint, mountpoint);
+        } else {
             fprintf(stderr, "fuse: invalid argument `%s'\n", arg);
             return -1;
         }
 
     default:
-    case KEY_KEEP:
         return 1;
     }
 }
@@ -165,7 +170,7 @@ int fuse_parse_cmdline(struct fuse_args *args, char **mountpoint,
     return -1;
 }
 
-static int fuse_daemonize(int foreground)
+int fuse_daemonize(int foreground)
 {
     int res;
 
@@ -186,15 +191,49 @@ static int fuse_daemonize(int foreground)
     return 0;
 }
 
+static struct fuse_chan *fuse_mount_common(const char *mountpoint,
+                                           struct fuse_args *args)
+{
+    struct fuse_chan *ch;
+    int fd = fuse_mount_compat25(mountpoint, args);
+    if (fd == -1)
+        return NULL;
+
+    ch = fuse_kern_chan_new(fd);
+    if (!ch)
+        fuse_kern_unmount(mountpoint, fd);
+
+    return ch;
+}
+
+struct fuse_chan *fuse_mount(const char *mountpoint, struct fuse_args *args)
+{
+    return fuse_mount_common(mountpoint, args);
+}
+
+static void fuse_unmount_common(const char *mountpoint, struct fuse_chan *ch)
+{
+    int fd = ch ? fuse_chan_fd(ch) : -1;
+    fuse_kern_unmount(mountpoint, fd);
+    fuse_chan_destroy(ch);
+}
+
+void fuse_unmount(const char *mountpoint, struct fuse_chan *ch)
+{
+    fuse_unmount_common(mountpoint, ch);
+}
+
 static struct fuse *fuse_setup_common(int argc, char *argv[],
                                       const struct fuse_operations *op,
                                       size_t op_size,
                                       char **mountpoint,
                                       int *multithreaded,
                                       int *fd,
+                                      void *user_data,
                                       int compat)
 {
     struct fuse_args args = FUSE_ARGS_INIT(argc, argv);
+    struct fuse_chan *ch;
     struct fuse *fuse;
     int foreground;
     int res;
@@ -203,67 +242,73 @@ static struct fuse *fuse_setup_common(int argc, char *argv[],
     if (res == -1)
         return NULL;
 
-    *fd = fuse_mount(*mountpoint, &args);
-    if (*fd == -1) {
+    ch = fuse_mount_common(*mountpoint, &args);
+    if (!ch) {
         fuse_opt_free_args(&args);
         goto err_free;
     }
 
-    fuse = fuse_new_common(*fd, &args, op, op_size, compat);
+    fuse = fuse_new_common(ch, &args, op, op_size, user_data, compat);
     fuse_opt_free_args(&args);
     if (fuse == NULL)
         goto err_unmount;
 
     res = fuse_daemonize(foreground);
     if (res == -1)
-        goto err_destroy;
+        goto err_unmount;
 
     res = fuse_set_signal_handlers(fuse_get_session(fuse));
     if (res == -1)
-        goto err_destroy;
+        goto err_unmount;
+
+    if (fd)
+        *fd = fuse_chan_fd(ch);
 
     return fuse;
 
- err_destroy:
-    fuse_destroy(fuse);
  err_unmount:
-    fuse_unmount(*mountpoint);
+    fuse_unmount_common(*mountpoint, ch);
+    if (fuse)
+        fuse_destroy(fuse);
  err_free:
     free(*mountpoint);
     return NULL;
 }
 
 struct fuse *fuse_setup(int argc, char *argv[],
-                        const struct fuse_operations *op,
-                        size_t op_size, char **mountpoint,
-                        int *multithreaded, int *fd)
+                        const struct fuse_operations *op, size_t op_size,
+                        char **mountpoint, int *multithreaded, void *user_data)
 {
     return fuse_setup_common(argc, argv, op, op_size, mountpoint,
-                             multithreaded, fd, 0);
+                             multithreaded, NULL, user_data, 0);
 }
 
-void fuse_teardown(struct fuse *fuse, int fd, char *mountpoint)
+static void fuse_teardown_common(struct fuse *fuse, char *mountpoint)
 {
-    (void) fd;
-
-    fuse_remove_signal_handlers(fuse_get_session(fuse));
-    fuse_unmount(mountpoint);
+    struct fuse_session *se = fuse_get_session(fuse);
+    struct fuse_chan *ch = fuse_session_next_chan(se, NULL);
+    fuse_remove_signal_handlers(se);
+    fuse_unmount_common(mountpoint, ch);
     fuse_destroy(fuse);
     free(mountpoint);
 }
 
+void fuse_teardown(struct fuse *fuse, char *mountpoint)
+{
+    fuse_teardown_common(fuse, mountpoint);
+}
+
 static int fuse_main_common(int argc, char *argv[],
                             const struct fuse_operations *op, size_t op_size,
-                            int compat)
+                            void *user_data, int compat)
 {
     struct fuse *fuse;
     char *mountpoint;
     int multithreaded;
     int res;
-    int fd;
 
     fuse = fuse_setup_common(argc, argv, op, op_size, &mountpoint,
-                             &multithreaded, &fd, compat);
+                             &multithreaded, NULL, user_data, compat);
     if (fuse == NULL)
         return 1;
 
@@ -272,7 +317,7 @@ static int fuse_main_common(int argc, char *argv[],
     else
         res = fuse_loop(fuse);
 
-    fuse_teardown(fuse, fd, mountpoint);
+    fuse_teardown_common(fuse, mountpoint);
     if (res == -1)
         return 1;
 
@@ -280,29 +325,30 @@ static int fuse_main_common(int argc, char *argv[],
 }
 
 int fuse_main_real(int argc, char *argv[], const struct fuse_operations *op,
-                   size_t op_size)
+                   size_t op_size, void *user_data)
 {
-    return fuse_main_common(argc, argv, op, op_size, 0);
+    return fuse_main_common(argc, argv, op, op_size, user_data, 0);
 }
 
 #undef fuse_main
+int fuse_main(void);
 int fuse_main(void)
 {
     fprintf(stderr, "fuse_main(): This function does not exist\n");
     return -1;
 }
 
-#ifndef __FreeBSD__
-
 #include "fuse_compat.h"
 
+#ifndef __FreeBSD__
+
 struct fuse *fuse_setup_compat22(int argc, char *argv[],
                                  const struct fuse_operations_compat22 *op,
                                  size_t op_size, char **mountpoint,
                                  int *multithreaded, int *fd)
 {
     return fuse_setup_common(argc, argv, (struct fuse_operations *) op,
-                             op_size, mountpoint, multithreaded, fd, 22);
+                             op_size, mountpoint, multithreaded, fd, NULL, 22);
 }
 
 struct fuse *fuse_setup_compat2(int argc, char *argv[],
@@ -312,7 +358,7 @@ struct fuse *fuse_setup_compat2(int argc, char *argv[],
 {
     return fuse_setup_common(argc, argv, (struct fuse_operations *) op,
                              sizeof(struct fuse_operations_compat2),
-                             mountpoint, multithreaded, fd, 21);
+                             mountpoint, multithreaded, fd, NULL, 21);
 }
 
 int fuse_main_real_compat22(int argc, char *argv[],
@@ -320,21 +366,28 @@ int fuse_main_real_compat22(int argc, char *argv[],
                             size_t op_size)
 {
     return fuse_main_common(argc, argv, (struct fuse_operations *) op, op_size,
-                            22);
+                            NULL, 22);
 }
 
 void fuse_main_compat1(int argc, char *argv[],
                       const struct fuse_operations_compat1 *op)
 {
     fuse_main_common(argc, argv, (struct fuse_operations *) op,
-                     sizeof(struct fuse_operations_compat1), 11);
+                     sizeof(struct fuse_operations_compat1), NULL, 11);
 }
 
 int fuse_main_compat2(int argc, char *argv[],
                       const struct fuse_operations_compat2 *op)
 {
     return fuse_main_common(argc, argv, (struct fuse_operations *) op,
-                            sizeof(struct fuse_operations_compat2), 21);
+                            sizeof(struct fuse_operations_compat2), NULL, 21);
+}
+
+int fuse_mount_compat1(const char *mountpoint, const char *args[])
+{
+    /* just ignore mount args for now */
+    (void) args;
+    return fuse_mount_compat22(mountpoint, NULL);
 }
 
 __asm__(".symver fuse_setup_compat2,__fuse_setup@");
@@ -344,3 +397,37 @@ __asm__(".symver fuse_main_compat2,fuse_main@");
 __asm__(".symver fuse_main_real_compat22,fuse_main_real@FUSE_2.2");
 
 #endif /* __FreeBSD__ */
+
+
+struct fuse *fuse_setup_compat25(int argc, char *argv[],
+                                 const struct fuse_operations_compat25 *op,
+                                 size_t op_size, char **mountpoint,
+                                 int *multithreaded, int *fd)
+{
+    return fuse_setup_common(argc, argv, (struct fuse_operations *) op,
+                             op_size, mountpoint, multithreaded, fd, NULL, 25);
+}
+
+int fuse_main_real_compat25(int argc, char *argv[],
+                            const struct fuse_operations_compat25 *op,
+                            size_t op_size)
+{
+    return fuse_main_common(argc, argv, (struct fuse_operations *) op, op_size,
+                            NULL, 25);
+}
+
+void fuse_teardown_compat22(struct fuse *fuse, int fd, char *mountpoint)
+{
+    (void) fd;
+    fuse_teardown_common(fuse, mountpoint);
+}
+
+int fuse_mount_compat25(const char *mountpoint, struct fuse_args *args)
+{
+    return fuse_kern_mount(mountpoint, args);
+}
+
+__asm__(".symver fuse_setup_compat25,fuse_setup@FUSE_2.5");
+__asm__(".symver fuse_teardown_compat22,fuse_teardown@FUSE_2.2");
+__asm__(".symver fuse_main_real_compat25,fuse_main_real@FUSE_2.5");
+__asm__(".symver fuse_mount_compat25,fuse_mount@FUSE_2.5");
